import numpy as np
import logging
import os
import argparse
import gensim
from gensim.models import word2vec
from gensim.models import KeyedVectors as kv

import extraction
import evaluation
import config

fname = config.EMBEDDINGS_PATH
print("Loading model.", end="\r")
try:
    model = kv.load(fname)
except: 
    model = kv.load_word2vec_format(fname, binary=False)
print("Done.         ")

def next_mention(mentions, anaphora):
    sims = [1/(anaphora.span[0] - mention.span[1]) if mention.span[1] < anaphora.span[0] else 0 for mention in mentions]
    return mentions[np.argmax(sims)]

def head_match(mentions, anaphora):
    # work in progress
    match_list = list()
    for mention in mentions:
        if mention.span[1] < anaphora.span[0]:
            overlap = len([m for m in mention.tokens if m in anaphora.tokens])
            overlap_ratio = 2*overlap/(len(mention.tokens)+len(anaphora.tokens))
            match_list.append(overlap_ratio)
        else:
            match_list.append(0)
    return mentions[np.argmax(match_list)]
        
def best_embedding(mentions, anaphora):
    sims = [1/(model.wv.wmdistance(mention.tokens, anaphora.tokens)+1) if mention.span[1] < anaphora.span[0] else 0 for mention in mentions]
    return mentions[np.argmax(sims)]

def most_salient_entity(entities, anaphora):
    lens = [len(entity) if entity[0].span[1] < anaphora.span[0] else 0 for entity in entities]
    return entities[np.argmax(lens)]

def largest_entity_span(entities, anaphora):
    spans = [entity[-1].span[1] - entity[0].span[0] if entity[0].span[1] < anaphora.span[0] else 0 for entity in entities]
    return entities[np.argmax(spans)]
    
if __name__ == "__main__":
    false = {"next_mention": 0, "best_embedding": 0, "head":0, "mse": 0, "les": 0}
    true = {"next_mention": 0, "best_embedding": 0, "head":0, "mse": 0, "les": 0}
    
    root = config.DATA_PATH
    files = os.listdir(root)
    
    for f in files[:]:
        
        print("Fetching test data...")
        mentions = extraction.from_isnotes(root+f)
        
        comparatives = [m for m in mentions if m.comp_from and m.comp_from != "outsidetext"]
        
        entities = []
        entity_dict = {}
        for mention in mentions:
            if not mention.coref_set:
                entities.append([mention])
            else:
                if mention.coref_set in entity_dict:
                    entity_dict[mention.coref_set].append(mention)
                else:
                    entity_dict[mention.coref_set] = [mention]
        
        entities += list(entity_dict.values())


        closest = lambda E, c: E[np.argmax([1/(c.span[0] - e.span[1]) if c.span[0] > e.span[1] else 0 for e in E])] 
        
        for comp in comparatives:
            
            if next_mention(mentions, comp).mention_id == comp.comp_from:
                true["next_mention"] += 1
            else:
                false["next_mention"] += 1
            if best_embedding(mentions, comp).mention_id == comp.comp_from:
                true["best_embedding"] += 1
            else:
                false["best_embedding"] += 1
            if head_match(mentions, comp).mention_id == comp.comp_from:
                true["head"] += 1
            else:
                false["head"] += 1
            
            mse_result = most_salient_entity(entities, comp)
            les_result = largest_entity_span(entities, comp)
            
            if closest(mse_result, comp).mention_id == comp.comp_from:
                true["mse"] += 1
            else:
                false["mse"] += 1
            if closest(les_result, comp).mention_id == comp.comp_from:
                true["les"] += 1
            else:
                false["les"] += 1
        
        recall = {"next_mention": true["next_mention"]/(true["next_mention"]+false["next_mention"]), "best_embedding": true["best_embedding"]/(true["best_embedding"]+false["best_embedding"]), "head": true["head"]/(true["head"]+false["head"]), "mse": true["mse"]/(true["mse"]+false["mse"]), "les": true["les"]/(true["les"]+false["les"])}
        
        print("Recall after {}:".format(f))
        print("next_mention:\t{}\t(True: {}, False:{})".format(recall["next_mention"], true["next_mention"], false["next_mention"]))
        print("best_embedding:\t{}\t(True: {}, False:{})".format(recall["best_embedding"], true["best_embedding"], false["best_embedding"]))
        print("head_match:\t{}\t(True: {}, False:{})".format(recall["head"], true["head"], false["head"]))
        print("most_salient_entity:\t{}\t(True: {}, False:{})".format(recall["mse"], true["mse"], false["mse"]))
        print("largest_entity_span:\t{}\t(True: {}, False:{})".format(recall["les"], true["les"], false["les"]))
