"""
Fetches all documents including comparative constructions.
"""

import os
import sys
import re

PATH = "/proj/zimmermann/comparatives/corpus/"
TARGET_PATH = "/proj/zimmermann/comparatives/annotation/OntoNotes/"
COMPARATIVES = ["other", "similar", "comparable", "different", "additional", "extra"]

def format():
    
    c = [0,0,0,0]
    
    for root, dirs, files in os.walk(PATH):
        for f in files:
            with open(os.path.join(root, f), "r") as document:
                current_word_id = 0
                markable_id = 0
                comp_id = -1
                set_id = 0
                
                jjr_count = 0
                rbr_count = 0
                jj_count = 0
                
                mmax = open(TARGET_PATH+f+".mmax",'w')
                mmax.write('<?xml version="1.0"?>\n<mmax_project>\n<sentences></sentences>\n<words>{}.xml</words>\n<gestures></gestures>\n<keyactions></keyactions>\n</mmax_project>'.format(f))
                mmax.close()
                
                xml = open(TARGET_PATH+"Basedata/"+f+".xml",'w')
                xml.write('<?xml version="1.0" encoding="US-ASCII"?>\n<!DOCTYPE words SYSTEM "words.dtd">\n<words>')
                
                comp_level = open(TARGET_PATH+"Markables/"+f+"_comparison_level.xml",'w')
                comp_level.write('<?xml version="1.0" encoding="UTF-8"?>\n<!DOCTYPE markables SYSTEM "markables.dtd">\n<markables xmlns="www.eml.org/NameSpaces/comparison">')
                
                coref_level = open(TARGET_PATH+"Markables/"+f+"_coref_level.xml",'w')
                coref_level.write('<?xml version="1.0" encoding="UTF-8"?>\n<!DOCTYPE markables SYSTEM "markables.dtd">\n<markables xmlns="www.eml.org/NameSpaces/coref">')
                
                
                
                last_pos = ""
                last_word = ""
                stack = list()
                coref_stack = dict()
                np_with_jj = False
                comp_word = None
                in_comp = 0
                buff="" #buffer saves annotation after * for next line
                
                for line in document:
                    if line.strip():
                        
                        if line.startswith("#"):
                            continue
                        
                        cols = line.strip().split()

                        xml.write('<word id="word_{}">{}</word>\n'.format(current_word_id,cols[3]))
                        
                        
                        synt = cols[5]
                        
                        new_tags = re.findall("[A-Z]+", synt)
                        open_brackets = synt.count("(")
                        closed_brackets = synt.count(")")
                        
                        stack+=new_tags
                        
                        if "NP" in new_tags:
                            np_start_word_id = current_word_id
                            np_with_jj = False
                        
                        if closed_brackets and "NP" in stack[-1*closed_brackets:] and in_comp and np_start_word_id <= comp_id <= current_word_id:
                            comp_level.write('<markable id="markable_{}" span="word_{}..word_{}" comparison_type="anaphora" comparison_class="set_{}" mmax_level="comparison"/>\n'.format(markable_id, np_start_word_id, current_word_id, set_id))
                            in_comp = False
                            markable_id += 1
                            set_id += 1
                    
                        
                        if cols[4] == "JJ" and cols[3] in COMPARATIVES:
                            np_with_jj = True
                            comp_id = current_word_id
                                
                        if cols[4] == "JJR" and "NP" in stack:
                            set_id += 1
                            comp_id = current_word_id
                            in_comp = True
                            c[0] += 1
                            comp_level.write('<markable id="markable_{}" span="word_{}" comparative_type="anaphoric"  comparison_type="comparative"  comparison_class="set_{}"  mmax_level="comparison" />\n'.format(markable_id, current_word_id, set_id))
                            markable_id += 1
                        elif cols[4] == "JJ" and last_pos == "RBR" and "NP" in stack:
                            set_id += 1
                            comp_id = current_word_id
                            in_comp = True
                            c[1] += 1
                            comp_level.write('<markable id="markable_{}" span="word_{}..word_{}" comparative_type="anaphoric"  comparison_type="comparative"  comparison_class="set_{}"  mmax_level="comparison" />\n'.format(markable_id, current_word_id-1, current_word_id ,set_id))
                            markable_id += 1
                        elif (cols[4] == "NN" or cols[4] == "NNS") and np_with_jj:
                            set_id += 1
                            in_comp = True
                            c[2] += 1
                            comp_level.write('<markable id="markable_{}" span="word_{}" comparative_type="anaphoric"  comparison_type="comparative"  comparison_class="set_{}"  mmax_level="comparison" />\n'.format(markable_id, comp_id, set_id))
                            markable_id += 1
                            np_with_jj = False
                        elif cols[3].lower() == "others":
                            set_id += 1
                            comp_id = current_word_id
                            c[3] += 1
                            comp_level.write('<markable id="markable_{}" span="word_{}" comparative_type="anaphoric"  comparison_type="comparative"  comparison_class="set_{}"  mmax_level="comparison" />\n'.format(markable_id, current_word_id, set_id))
                            comp_level.write('<markable id="markable_{}" span="word_{}" comparison_type="anaphora" comparison_class="set_{}" mmax_level="comparison"/>\n'.format(markable_id, current_word_id, set_id))
                            markable_id += 1
                            set_id += 1
                            
                        last_pos = cols[4]
                        last_word = cols[3]
                        
                        if closed_brackets > 0:
                            del stack[-1*closed_brackets:]
                        
                        
                        coref_anno = cols[-1]
                        
                        for coref in coref_anno.split("|"):
                            start = False
                            end = False
                            buffr = ""
                            for i, char in enumerate(coref):
                                if char == "(":
                                    start = True
                                elif char == ")":
                                    end = True
                                elif char in "0123456789":
                                    buffr += char
                            if start and end:
                                coref_level.write('<markable id="markable_{}" span="word_{}" agreement="none"  np_form="none"  coref_class="set_{}"  mmax_level="coref"  semantic_class="none"  type="none"  grammatical_role="none" />'.format(markable_id, current_word_id, buffr))
                                markable_id += 1
                            elif start:
                                if buffr not in coref_stack:
                                    coref_stack[buffr] = [current_word_id]
                                else:
                                    coref_stack[buffr].append(current_word_id)
                            elif end:
                                coref_level.write('<markable id="markable_{}" span="word_{}..word_{}" agreement="none"  np_form="none"  coref_class="set_{}"  mmax_level="coref"  semantic_class="none"  type="none"  grammatical_role="none" />'.format(markable_id, coref_stack[buffr][-1], current_word_id, buffr))
                                markable_id += 1
                        
                        current_word_id += 1
                        
                xml.write("</words>")
                xml.close()
                comp_level.write("</markables>")
                comp_level.close()
                coref_level.write("</markables>")
                coref_level.close()
                
        print(c)
    
if __name__ == "__main__":
    format()
